def get(name, task_config):
    subtree = task_config.subtree(r'^uda\.%s\.' % name, walk_subtree=False)
    if 'uda' in subtree and name in subtree['uda']:
        return subtree['uda'][name]
    return None

def get_configured(task_config):
    subtree = task_config.subtree(r'^uda\.', walk_subtree=False)
    if 'uda' in subtree:
        return {k: v['type'] if 'type' in v else 'string' for k, v in subtree['uda'].items()}
    return {}



from functools import reduce

from vit.formatter import String
from vit.util import unicode_len

class Description(String):
    def format(self, description, task):
        if not description:
            return self.empty()
        width = unicode_len(description)
        colorized_description = self.colorize_description(description)
        if task['annotations']:
            annotation_width, colorized_description = self.format_combined(colorized_description, task)
            if annotation_width > width:
                width = annotation_width
        return (width, colorized_description)

    def format_description_truncated(self, description):
        return '%s...' % description[:self.formatter.description_truncate_len] if unicode_len(description) > self.formatter.description_truncate_len else description

    def format_combined(self, colorized_description, task):
        annotation_width, formatted_annotations = self.format_annotations(task)
        return annotation_width, colorized_description + [(None, "\n"), (None, formatted_annotations)]

    def format_annotations(self, task):
        def reducer(accum, annotation):
            width, formatted_list = accum
            formatted = self.format_annotation(annotation)
            new_width = unicode_len(formatted)
            if new_width > width:
                width = new_width
            formatted_list.append(formatted)
            return (width, formatted_list)
        width, formatted_annotations = reduce(reducer, task['annotations'], (0, []))
        return width, "\n".join(formatted_annotations)

    def format_annotation(self, annotation):
        return '  %s %s' % (annotation['entry'].strftime(self.formatter.annotation), annotation['description'])

    def colorize(self, part):
        return self.colorizer.keyword(part)

    def colorize_description(self, description):
        first_part, rest = self.colorizer.extract_keyword_parts(description)
        if first_part is None:
            return [(None, description)]
        def reducer(accum, part):
            if part:
                last_color, last_part = accum[-1]
                color, part = self.markup_element(part)
                if color == last_color:
                    accum[-1] = (last_color, last_part + part)
                    return accum
                else:
                    return accum + [(color, part)]
            return accum
        return reduce(reducer, rest, [self.markup_element(first_part)])



from vit.formatter.depends import Depends

class DependsIndicator(Depends):
    def format_list(self, depends, task):
        return self.formatter.indicator_dependency if self.filter_by_blocking_task_uuids(depends) else ''



from vit.formatter import String

class Status(String):
    def markup_element(self, status):
        return (self.colorize(status), self.status_format(status))

    def colorize(self, status):
        return self.colorizer.status(status)

    def status_format(self, status):
        return status.capitalize()



import string
import re

class Readline:
    def __init__(self, edit_obj):
        self.edit_obj = edit_obj
        word_chars = string.ascii_letters + string.digits + "_"
        self._word_regex1 = re.compile(
            "([%s]+)" % "|".join(re.escape(ch) for ch in word_chars)
        )
        self._word_regex2 = re.compile(
            "([^%s]+)" % "|".join(re.escape(ch) for ch in word_chars)
        )

    def keys(self):
        return ('ctrl p', 'ctrl n', 'ctrl a', 'ctrl e', 'ctrl b', 'ctrl f',
                'ctrl h', 'ctrl d', 'ctrl t', 'ctrl u', 'ctrl k', 'meta b',
                'meta f', 'ctrl w', 'meta d')

    def keypress(self, key):
        # Move to the previous line.
        if key in ('ctrl p',):
            text = self.edit_obj.history.previous(self.edit_obj.metadata['history'])
            if text != False:
                self.edit_obj.set_edit_text(text)
            return None
        # Move to the next line.
        elif key in ('ctrl n',):
            text = self.edit_obj.history.next(self.edit_obj.metadata['history'])
            if text != False:
                self.edit_obj.set_edit_text(text)
            return None
        # Jump to the beginning of the line.
        elif key in ('ctrl a',):
            self.edit_obj.set_edit_pos(0)
            return None
        # Jump to the end of the line.
        elif key in ('ctrl e',):
            self.edit_obj.set_edit_pos(len(self.edit_obj.get_edit_text()))
            return None
        # Jump backward one character.
        elif key in ('ctrl b',):
            self.edit_obj.set_edit_pos(self.edit_obj.edit_pos - 1)
            return None
        # Jump forward one character.
        elif key in ('ctrl f',):
            self.edit_obj.set_edit_pos(self.edit_obj.edit_pos + 1)
            return None
        # Delete previous character.
        elif key in ('ctrl h',):
            if self.edit_obj.edit_pos > 0:
                self.edit_obj.set_edit_pos(self.edit_obj.edit_pos - 1)
                self.edit_obj.set_edit_text(
                    self.edit_obj.get_edit_text()[0 : self.edit_obj.edit_pos]
                    + self.edit_obj.get_edit_text()[self.edit_obj.edit_pos + 1 :])
            return None
        # Delete next character.
        elif key in ('ctrl d',):
            if self.edit_obj.edit_pos < len(self.edit_obj.get_edit_text()):
                edit_pos = self.edit_obj.edit_pos
                self.edit_obj.set_edit_text(
                    self.edit_obj.get_edit_text()[0 : self.edit_obj.edit_pos] +
                    self.edit_obj.get_edit_text()[self.edit_obj.edit_pos + 1 :])
                self.edit_obj.set_edit_pos(edit_pos)
            return None
        # Transpose characters.
        elif key in ('ctrl t',):
            # Can't transpose if there are less than 2 chars
            if len(self.edit_obj.get_edit_text()) < 2:
                return None
            self.edit_obj.set_edit_pos(max(2, self.edit_obj.edit_pos + 1))
            new_edit_pos = self.edit_obj.edit_pos

            edit_text = self.edit_obj.get_edit_text()
            self.edit_obj.set_edit_text(
                edit_text[: new_edit_pos - 2] +
                edit_text[new_edit_pos - 1] +
                edit_text[new_edit_pos - 2] +
                edit_text[new_edit_pos :])
            self.edit_obj.set_edit_pos(new_edit_pos)
            return None
        # Delete backwards to the beginning of the line.
        elif key in ('ctrl u',):
            self.edit_obj.set_edit_text(self.edit_obj.get_edit_text()[self.edit_obj.edit_pos :])
            self.edit_obj.set_edit_pos(0)
            return None
        # Delete forwards to the end of the line.
        elif key in ('ctrl k',):
            self.edit_obj.set_edit_text(self.edit_obj.get_edit_text()[: self.edit_obj.edit_pos])
            return None
        # Jump backward one word.
        elif key in ('meta b',):
            self.jump_backward_word()
            return None
        # Jump forward one word.
        elif key in ('meta f',):
            self.jump_forward_word()
            return None
        # Delete backwards to the beginning of the current word.
        elif key in ('ctrl w',):
            old_edit_pos = self.edit_obj.edit_pos
            self.jump_backward_word()
            new_edit_pos = self.edit_obj.edit_pos
            self.edit_obj.set_edit_text(self.edit_obj.edit_text[: new_edit_pos]
                + self.edit_obj.edit_text[old_edit_pos :])
            self.edit_obj.set_edit_pos(new_edit_pos)
            return None
        # Delete forwards to the end of the current word.
        elif key in ('meta d',):
            edit_pos = self.edit_obj.edit_pos
            self.jump_forward_word()
            self.edit_obj.set_edit_text(self.edit_obj.edit_text[: edit_pos]
                + self.edit_obj.edit_text[self.edit_obj.edit_pos :])
            self.edit_obj.set_edit_pos(edit_pos)
            return None

    def jump_backward_word(self):
        for match in self._word_regex1.finditer(
            self.edit_obj.edit_text[: self.edit_obj.edit_pos][::-1]
        ):
            self.edit_obj.set_edit_pos(self.edit_obj.edit_pos - match.end(1))
            return
        self.edit_obj.set_edit_pos(0)

    def jump_forward_word(self):
        for match in self._word_regex2.finditer(
            self.edit_obj.edit_text[self.edit_obj.edit_pos :]
        ):
            if match.start(1) > 0:
                self.edit_obj.set_edit_pos(self.edit_obj.edit_pos + match.start(1))
                return
        self.edit_obj.set_edit_pos(len(self.edit_obj.edit_text))




from vit.formatter import DateTime

class Scheduled(DateTime):
    def get_scheduled_state(self, scheduled, task):
        return self.formatter.get_scheduled_state(scheduled, task)

    def colorize(self, scheduled, task):
        return self.colorizer.scheduled(self.get_scheduled_state(scheduled, task))



from vit.formatter.description import Description
from vit.util import unicode_len

class DescriptionCount(Description):
    def format(self, description, task):
        if not description:
            return self.empty()
        width = unicode_len(description)
        colorized_description = self.colorize_description(description)
        if not task['annotations']:
            return (width, colorized_description)
        else:
            count_width, colorized_description = self.format_count(colorized_description, task)
            return (width + count_width, colorized_description)

    def format_count(self, colorized_description, task):
        count_string = self.format_annotation_count(task)
        return unicode_len(count_string), colorized_description + [(None, count_string)]

    def format_annotation_count(self, task):
        return " [%d]" % len(task['annotations'])



from vit import util
from vit.formatter.project import Project
from vit.util import unicode_len

class ProjectParent(Project):
    def format(self, project, task):
        parent = util.project_get_root(project)
        return (unicode_len(parent), self.markup_element(parent)) if parent else self.markup_none(self.colorizer.project_none())




from vit.formatter.description_count import DescriptionCount
from vit.util import unicode_len

class DescriptionTruncatedCount(DescriptionCount):
    def format(self, description, task):
        if not description:
            return self.empty()
        truncated_description = self.format_description_truncated(description)
        width = unicode_len(truncated_description)
        colorized_description = self.colorize_description(truncated_description)
        if not task['annotations']:
            return (width, colorized_description)
        else:
            count_width, colorized_description = self.format_count(colorized_description, task)
            return (width + count_width, colorized_description)



from vit.formatter.description import Description

class DescriptionOneline(Description):
    def format_combined(self, colorized_description, task):
        formatted_annotations = self.format_annotations(task)
        return 0, colorized_description + [(None, formatted_annotations)]

    def format_annotations(self, task):
        formatted_annotations = [self.format_annotation(annotation) for annotation in task['annotations']]
        return "".join(formatted_annotations)

    def format_annotation(self, annotation):
        return ' %s %s' % (annotation['entry'].strftime(self.formatter.annotation), annotation['description'])



import re

import urwid

BRACKETS_REGEX = re.compile("[<>]")

class BaseListBox(urwid.ListBox):
    """Maps task list shortcuts to default ListBox class.
    """

    def __init__(self, body, event=None, request_reply=None, action_manager=None):
        self.previous_focus_position = None
        self.list_walker = body
        self.event = event
        self.request_reply = request_reply
        self.action_manager = action_manager
        self.key_cache = self.request_reply.request('application:key_cache')
        self.register_managed_actions()
        return super().__init__(body)

    def get_top_middle_bottom_rows(self, size):
        #try:
            if len(self.list_walker) == 0:
                return None, None, None
            ((_, focused, _, _, _), (_, top_list), (_, bottom_list)) = self.calculate_visible(size)
            top = top_list[len(top_list) - 1][0] if len(top_list) > 0 else None
            bottom = bottom_list[len(bottom_list) - 1][0] if len(bottom_list) > 0 else None
            top_list_reversed = []
            # Neither top_list.reverse() nor reversed(top_list) works here, WTF?
            while True:
                if len(top_list) > 0:
                    row = top_list.pop()
                    top_list_reversed.append(row)
                else:
                    break
            assembled_list = top_list_reversed + [(focused, )] + (bottom_list if bottom else [])
            middle_list_position = (len(assembled_list) // 2) + 1
            middle_list = assembled_list[:middle_list_position] if len(assembled_list) > 1 else assembled_list
            middle = middle_list.pop()[0]
            return (top or focused), middle, (bottom or focused)
        #except:
        #    # TODO: Log this?
        #    return None, None, None

    def register_managed_actions(self):
        self.action_manager_registrar = self.action_manager.get_registrar()
        self.action_manager_registrar.register('LIST_UP', self.keypress_up)
        self.action_manager_registrar.register('LIST_DOWN', self.keypress_down)
        self.action_manager_registrar.register('LIST_PAGE_UP', self.keypress_page_up)
        self.action_manager_registrar.register('LIST_PAGE_DOWN', self.keypress_page_down)
        self.action_manager_registrar.register('LIST_HOME', self.keypress_home)
        self.action_manager_registrar.register('LIST_END', self.keypress_end)
        self.action_manager_registrar.register('LIST_SCREEN_TOP', self.keypress_screen_top)
        self.action_manager_registrar.register('LIST_SCREEN_MIDDLE', self.keypress_screen_middle)
        self.action_manager_registrar.register('LIST_SCREEN_BOTTOM', self.keypress_screen_bottom)
        self.action_manager_registrar.register('LIST_FOCUS_VALIGN_CENTER', self.keypress_focus_valign_center)

    # NOTE: The non-standard key presses work around infinite recursion while
    #       allowing the up, down, page up, and page down keys to be controlled
    #       from the keybinding file.
    def keypress_up(self, size):
        self.keypress(size, '<Up>')

    def keypress_down(self, size):
        self.keypress(size, '<Down>')

    def keypress_page_up(self, size):
        self.keypress(size, '<Page Up>')

    def keypress_page_down(self, size):
        self.keypress(size, '<Page Down>')

    def keypress_home(self, size):
        if len(self.body) > 0:
            self.set_focus(0)

    def keypress_end(self, size):
        if len(self.body) > 0:
            self.set_focus(len(self.body) - 1)
            self.set_focus_valign('bottom')

    def keypress_screen_top(self, size):
        top, _, _ = self.get_top_middle_bottom_rows(size)
        if top:
            self.set_focus(top.position)

    def keypress_screen_middle(self, size):
        _, middle, _ = self.get_top_middle_bottom_rows(size)
        if middle:
            self.set_focus(middle.position)

    def keypress_screen_bottom(self, size):
        _, _, bottom = self.get_top_middle_bottom_rows(size)
        if bottom:
            self.set_focus(bottom.position)

    def keypress_focus_valign_center(self, size):
        if len(self.body) > 0:
            self.set_focus(self.focus_position)
            self.set_focus_valign('middle')

    def transform_special_keys(self, key):
        # NOTE: These are special key presses passed to allow navigation
        #       keys to be managed via keybinding configuration. They are
        #       converted back to standard key presses here.
        if key in ['<Up>', '<Down>', '<Page Up>', '<Page Down>']:
            key = re.sub(BRACKETS_REGEX, '', key).lower()
        return key

    def list_action_executed(self, size, key):
        pass

    def eat_other_keybindings(self):
        return False

    def keypress(self, size, key):
        keys = self.key_cache.get(key)
        if self.action_manager_registrar.execute_handler(keys, size):
            self.list_action_executed(size, key)
            return None
        if self.eat_other_keybindings() and self.key_cache.is_keybinding(keys):
            return None
        key = self.transform_special_keys(key)
        return super().keypress(size, key)





import os
from functools import reduce

import tasklib
from tasklib.task import Task
from tasklib.backends import TaskWarriorException

from vit import util
from vit.exception import VitException

class TaskListModel:
    def __init__(self, task_config, reports, report=None, data_location=None):

        if not data_location:
            data_location = task_config.subtree('data.location')
        self.data_location = os.path.expanduser(data_location)
        self.tw = tasklib.TaskWarrior(self.data_location)
        self.reports = reports
        self.report = report
        self.tasks = []
        if report:
            self.update_report(report)

    def add(self, contact):
        pass

    def active_report(self):
        return self.reports[self.report]

    def parse_error(self, err):
        messages = filter(lambda l : l.startswith('Error:'), str(err).splitlines())
        return "\n".join(messages)

    def update_report(self, report, context_filters=[], extra_filters=[]):
        self.report = report
        active_report = self.active_report()
        report_filters = active_report['filter'] if 'filter' in active_report else []
        filters = self.build_task_filters(context_filters, report_filters, extra_filters)
        try:
            self.tasks = self.tw.tasks.filter(filters) if filters else self.tw.tasks.all()
            # NOTE: tasklib uses lazy loading and some operation is necessary
            # for self.tasks to actually be populated here.
            # See https://github.com/robgolding/tasklib/issues/81
            len(self.tasks)
        except TaskWarriorException as err:
            raise VitException(self.parse_error(err))

    def build_task_filters(self, *all_filters):
        def reducer(accum, filters):
            if filters:
                accum.append('( %s )' % ' '.join(filters))
            return accum
        filter_parts = reduce(reducer, all_filters, [])
        return ' '.join(filter_parts) if filter_parts else ''

    def get_task(self, uuid):
        try:
            return self.tw.tasks.get(uuid=uuid)
        except Task.DoesNotExist:
            return False

    def task_id(self, uuid):
        try:
            task = self.tw.tasks.get(uuid=uuid)
            return util.task_id_or_uuid_short(task)
        except Task.DoesNotExist:
            return False

    def task_description(self, uuid, description):
        task = self.get_task(uuid)
        if task:
            task['description'] = description
            task.save()
            return task
        return False

    def task_annotate(self, uuid, description):
        task = self.get_task(uuid)
        if task:
            task.add_annotation(description)
            return task
        return False

    def task_denotate(self, uuid, annotation):
        task = self.get_task(uuid)
        if task:
            task.remove_annotation(annotation)
            return task
        return False

    def task_priority(self, uuid, priority):
        task = self.get_task(uuid)
        if task:
            task['priority'] = priority
            task.save()
            return task
        return False

    def task_project(self, uuid, project):
        task = self.get_task(uuid)
        if task:
            task['project'] = project
            task.save()
            return task
        return False

    def task_done(self, uuid):
        task = self.get_task(uuid)
        if task:
            try:
                task.done()
                return True, task
            except (Task.CompletedTask, Task.DeletedTask) as e:
                return False, e
        return False, None

    def task_delete(self, uuid):
        task = self.get_task(uuid)
        if task:
            try:
                task.delete()
                return True, task
            except Task.DeletedTask as e:
                return False, e
        return False, None

    def task_start_stop(self, uuid):
        task = self.get_task(uuid)
        if task:
            try:
                task.stop() if task.active else task.start()
                return True, task
            except (Task.CompletedTask, Task.DeletedTask, Task.ActiveTask, Task.InactiveTask) as e:
                return False, e
        return False, None

    def task_tags(self, uuid, tags):
        task = self.get_task(uuid)
        if task:
            for tag in tags:
                marker = tag[0]
                if marker in ('+', '-'):
                    tag = tag[1:]
                    if marker == '+':
                        task['tags'].add(tag)
                    elif tag in task['tags']:
                        task['tags'].remove(tag)
                else:
                    task['tags'].add(tag)
            task.save()
            return task
        return False

#    def get_summary(self, report=None):
#        report = report or self.report
#        self.update_report(report)
#        summary = []
#        for task in self.tasks:
#            summary.append(self.build_task_row(task))
#        return summary
#
#    def _reload_list(self, new_value=None):
#        self._list_view.options = self._model.get_summary()
#        self._list_view.value = new_value



from vit.formatter.tags import Tags

class TagsCount(Tags):
    def format(self, tags, task):
        if not tags:
            return self.markup_none(self.colorizer.tag_none())
        elif len(tags) == 1:
            return (3, (self.colorizer.tag(list(tags)[0]), '[1]'))
        else:
            tag_length = len(tags)
            indicator = '[%d]' % tag_length
            return (tag_length + 2, (self.colorizer.tag(''), indicator))



from operator import itemgetter
from collections import OrderedDict
from itertools import repeat
from functools import partial, reduce
from time import sleep
import re
import math
from functools import cmp_to_key

import urwid

from vit import util
from vit.base_list_box import BaseListBox
from vit.list_batcher import ListBatcher
from vit.formatter.project import Project as ProjectFormatter
from vit.util import unicode_len


REDUCE_COLUMN_WIDTH_LIMIT = 20
COLUMN_PADDING = 2
MARKER_COLUMN_NAME = 'markers'

class TaskTable:

    def __init__(self, config, task_config, formatter, screen, on_select=None, event=None, action_manager=None, request_reply=None, markers=None, draw_screen_callback=None):
        self.config = config
        self.task_config = task_config
        self.formatter = formatter
        self.screen = screen
        self.on_select = on_select
        self.event = event
        self.action_manager = action_manager
        self.request_reply = request_reply
        self.markers = markers
        self.draw_screen = draw_screen_callback
        self.list_walker = urwid.SimpleFocusListWalker([])
        self.row_striping = self.config.row_striping_enabled
        self.listbox = TaskListBox(self.list_walker, self.screen, event=self.event, request_reply=self.request_reply, action_manager=self.action_manager)
        self.init_event_listeners()
        self.set_request_callbacks()

    def init_event_listeners(self):
        def signal_handler():
            self.update_focus()
        urwid.connect_signal(self.list_walker, 'modified', signal_handler)
        def task_list_keypress(data):
            self.update_header(data['size'])
        self.event.listen('task-list:keypress', task_list_keypress)
        self.event.listen('task-list:size:change', self.size_changed)
        self.event.listen('task-list:keypress:down', self.task_list_keypress_down)
        self.event.listen('task-list:keypress:page_down', self.task_list_keypress_page_down)
        self.event.listen('task-list:keypress:end', self.task_list_keypress_end)
        self.event.listen('task-list:keypress:focus_valign_center', self.task_list_keypress_focus_valign_center)

    def set_request_callbacks(self):
        self.request_reply.set_handler('task-table:batch:next', 'Render next batch of tasks', lambda: self.get_next_task_batch())

    def get_next_task_batch(self):
        return self.batcher.add()

    def task_list_keypress_down(self, size):
        self.batcher.add(1)

    def task_list_keypress_page_down(self, size):
        _, rows = size
        self.batcher.add(rows)

    def task_list_keypress_end(self, size):
        self.batcher.add(0)

    def task_list_keypress_focus_valign_center(self, size):
        _, rows = size
        half_rows = math.ceil(rows / 2)
        self.batcher.add(half_rows)

    def get_blocking_task_uuids(self):
        return self.request_reply.request('application:blocking_task_uuids')

    def update_data(self, report, tasks):
        self.report = report
        self.tasks = tasks
        self.list_walker.clear()
        self.columns = []
        self.column_names = []
        self.rows = []
        self.sort()
        self.set_column_metadata()
        if self.markers.enabled:
            self.set_marker_columns()
            self.add_markers_column()
        self.indent_subprojects = self.subproject_indentable()
        self.project_cache = {}
        # TODO: This is for the project placeholders, feels sloppy.
        self.project_formatter = ProjectFormatter('project', self.report, self.formatter, self.get_blocking_task_uuids())
        self.build_rows()
        self.clean_columns()
        self.project_column_idx = self.get_project_column_idx()
        self.reconcile_column_width_for_label()
        self.resize_columns()
        self.build_table()
        self.listbox.set_focus_position()
        self.update_focus()

    def update_header(self, size):
        if self.project_column_idx is not None:
            self.update_project_column_header(size)

    def get_project_column_idx(self):
        for idx, column in enumerate(self.columns):
            if column['name'] == 'project':
                return idx
        return None

    def get_project_from_row(self, row):
        return row.task['project'] if isinstance(row, SelectableRow) else row.project

    def update_project_column_header(self, size):
        if self.indent_subprojects:
            top, _, _ = self.listbox.get_top_middle_bottom_rows(size)
            if top:
                project = self.get_project_from_row(top)
                if project:
                    _, parents = util.project_get_subproject_and_parents(project)
                    self.set_project_column_header(parents)
                else:
                    self.set_project_column_header()

    def set_project_column_header(self, parents=None):
        column_index = self.project_column_idx
        (columns_widget, _) = self.header.original_widget.contents[column_index]
        (widget, _) = columns_widget.contents[0]
        label = self.project_label_for_parents(parents)
        widget.original_widget.original_widget.set_text(label)

    def project_label_for_parents(self, parents):
        return '.'.join(parents) if parents else self.task_config.get_column_label(self.report['name'], 'project')

    def update_focus(self):
        if self.listbox.focus:
            if self.listbox.previous_focus_position != self.listbox.focus_position:
                if self.listbox.previous_focus_position is not None and self.listbox.previous_focus_position < len(self.list_walker):
                    self.list_walker[self.listbox.previous_focus_position].reset_attr_map()
                if self.listbox.focus_position is not None:
                    self.update_focus_attr('reveal focus')
                self.listbox.previous_focus_position = self.listbox.focus_position
            else:
                self.update_focus_attr('reveal focus')
        else:
            self.listbox.previous_focus_position = None

    def update_focus_attr(self, attr, position=None):
        if position is None:
            position = self.listbox.focus_position
        self.list_walker[position].row.set_attr_map({None: attr})

    def flash_focus(self, repeat_times=None, pause_seconds=None):
        if repeat_times is None:
            repeat_times = self.config.get_flash_focus_repeat_times()
        if pause_seconds is None:
            pause_seconds = self.config.get_flash_focus_pause_seconds()
        if self.listbox.focus:
            position = self.listbox.focus_position if self.listbox.focus_position is not None else self.listbox.previous_focus_position if self.listbox.previous_focus_position is not None else None
            if position is not None and repeat_times > 0:
                self.update_focus_attr('flash on', position)
                self.draw_screen()
                for i in repeat(None, repeat_times):
                    sleep(pause_seconds)
                    self.update_focus_attr('flash off', position)
                    self.draw_screen()
                    sleep(pause_seconds)
                    self.update_focus_attr('flash on', position)
                    self.draw_screen()
                sleep(pause_seconds)
                self.update_focus_attr('reveal focus', position)
                self.draw_screen()

    def sort(self):
        if 'sort' in self.report:
            for column, order, collate in reversed(self.report['sort']):
                def comparator_default(first, second):
                    if first[column] is not None and second[column] is not None:
                        return -1 if first[column] < second[column] else 1 if first[column] > second[column] else 0
                    elif first[column] is None and second[column] is None:
                        return 0
                    elif first[column] is not None and second[column] is None:
                        return -1
                    elif first[column] is None and second[column] is not None:
                        return 1

                def comparator_uda_string(values_index, first, second):
                    a, b = first[column], second[column]
                    default_index = len(values_index)
                    a_index = values_index.get(a or '', default_index)
                    b_index = values_index.get(b or '', default_index)
                    return (a_index > b_index) - (a_index < b_index)

                uda = self.task_config.uda_config.get(column)
                if uda is not None and uda.get('type') == 'string':
                    values_index = uda.get('values_index')
                    comparator = partial(comparator_uda_string, values_index)
                else:
                    comparator = comparator_default
                if order and order == 'descending':
                    self.tasks = sorted(self.tasks, key=cmp_to_key(comparator), reverse=True)
                else:
                    self.tasks = sorted(self.tasks, key=cmp_to_key(comparator))

    def column_formatter_kwargs(self):
        kwargs = {}
        if 'dateformat' in self.report:
            kwargs['custom_formatter'] = self.report['dateformat']
        return kwargs

    def set_marker_columns(self):
        self.report_marker_columns = [c for c in self.markers.columns if c not in self.column_names]

    def add_markers_column(self):
        name, formatter_class = self.formatter.get(MARKER_COLUMN_NAME)
        self.columns.insert(0, {
            'name': name,
            'label': self.markers.header_label,
            'formatter': formatter_class(self.report, self.formatter, self.report_marker_columns, self.get_blocking_task_uuids()),
            'width': 0,
            'align': 'right',
        })
        self.column_names.insert(0, name)

    def add_column(self, name, label, formatter_class, align='left'):
        self.columns.append({
            'name': name,
            'label': label,
            'formatter': formatter_class,
            'width': 0,
            'align': align,
        })
        self.column_names.append(name)

    def set_column_metadata(self):
        kwargs = self.column_formatter_kwargs()
        for idx, column_formatter in enumerate(self.report['columns']):
            name, formatter_class = self.formatter.get(column_formatter)
            self.add_column(name, self.report['labels'][idx], formatter_class(name, self.report, self.formatter, self.get_blocking_task_uuids(), **kwargs))

    def is_marker_column(self, column):
        return column == MARKER_COLUMN_NAME

    def has_marker_column(self):
        return MARKER_COLUMN_NAME in self.column_names

    def build_rows(self):
        self.task_row_striping_reset()
        for task in self.tasks:
            row_data = []
            self.inject_project_placeholders(task)
            alt_row = self.task_row_striping()
            for idx, column in enumerate(self.columns):
                formatted_value = column['formatter'].format(task[column['name']], task)
                width, text_markup = self.build_row_column(formatted_value)
                self.update_column_width(idx, column['width'], width)
                row_data.append(text_markup)
            self.rows.append(TaskRow(task, row_data, alt_row))

    def update_column_width(self, idx, current_width, new_width):
        if new_width > current_width:
            self.columns[idx]['width'] = new_width

    def build_row_column(self, formatted_value):
        if isinstance(formatted_value, tuple):
            return formatted_value
        else:
            width = unicode_len(formatted_value) if formatted_value else 0
            return width, formatted_value

    def subproject_indentable(self):
        return self.config.subproject_indentable and self.report['subproject_indentable']

    def inject_project_placeholders(self, task):
        project = task['project']
        if self.indent_subprojects and project:
            parents = self.project_may_need_placeholders(project)
            if parents:
                to_inject = self.build_project_placeholders_to_inject(parents, [])
                for project_parts in to_inject:
                    self.inject_project_placeholder(project_parts)

    def build_project_placeholders_to_inject(self, parents, to_inject):
        project = '.'.join(parents)
        if project in self.project_cache:
            return to_inject
        else:
            self.project_cache[project] = True
            to_inject.append(parents.copy())
            parents.pop()
            if len(parents) > 0:
                return self.build_project_placeholders_to_inject(parents, to_inject)
            else:
                to_inject.reverse()
                return to_inject

    def project_may_need_placeholders(self, project):
        if project not in self.project_cache:
            self.project_cache[project] = True
            _, parents = util.project_get_subproject_and_parents(project)
            return parents

    def inject_project_placeholder(self, project_parts):
        project = '.'.join(project_parts)
        (width, spaces, indicator, subproject) = self.formatter.format_subproject_indented(project_parts)
        # TODO: This is pretty ugly...
        alt_row = self.task_row_striping()
        self.rows.append(ProjectRow(project, [spaces, indicator, (self.project_formatter.colorize(project), subproject)], alt_row))

    def clean_columns(self):
        self.clean_markers_column() if self.task_config.print_empty_columns else self.clean_empty_columns()

    def clean_markers_column(self):
        self.non_filtered_columns = [False if (c['name'] == MARKER_COLUMN_NAME and c['width'] == 0) else c for c in self.columns]
        self.columns = [c for c in self.columns if not (c['name'] == MARKER_COLUMN_NAME and c['width'] == 0)]

    def clean_empty_columns(self):
        self.non_filtered_columns = [c if c['width'] > 0 else False for c in self.columns]
        self.columns = [c for c in self.columns if c['width'] > 0]

    def resize_columns(self):
        cols, _ = self.listbox.size
        padding = (len(self.columns) - 1) * COLUMN_PADDING
        total_width = padding
        to_adjust= []
        for idx, column in enumerate(self.columns):
            width = column['width']
            total_width += width
            if width > REDUCE_COLUMN_WIDTH_LIMIT:
                to_adjust.append({'idx': idx, 'width': width})
        if total_width > cols:
            self.adjust_oversized_columns(total_width - cols, to_adjust)
            if to_adjust:
                # This is called recursively to account for cases when further
                # reduction is necessary because one or more column's reductions
                # were limited to REDUCE_COLUMN_WIDTH_LIMIT.
                self.resize_columns()

    def adjust_oversized_columns(self, reduce_by, to_adjust):
        to_adjust = list(map(lambda c: c.update({'ratio': (c['width'] - REDUCE_COLUMN_WIDTH_LIMIT) / c['width']}) or c, to_adjust))
        ratio_total = reduce(lambda acc, c: acc + c['ratio'], to_adjust, 0)
        to_adjust = list(map(lambda c: c.update({'percentage': c['ratio'] / ratio_total}) or c, to_adjust))
        for c in to_adjust:
            adjusted_width = c['width'] - math.ceil(reduce_by * c['percentage'])
            self.columns[c['idx']]['width'] = adjusted_width if adjusted_width > REDUCE_COLUMN_WIDTH_LIMIT else REDUCE_COLUMN_WIDTH_LIMIT

    def reconcile_column_width_for_label(self):
        for idx, column in enumerate(self.columns):
            label_len = unicode_len(column['label'])
            if column['width'] < label_len:
                self.columns[idx]['width'] = label_len

    def get_alt_row_background_modifier(self):
        return '.striped-table-row'

    def task_row_striping_reset(self):
        self.task_alt_row = False

    def task_row_striping(self):
        if self.row_striping:
            self.task_alt_row = not self.task_alt_row
            modifier = self.task_alt_row and self.get_alt_row_background_modifier() or ''
            self.formatter.task_colorizer.set_background_modifier(modifier)
        return self.task_alt_row

    def format_task_batch(self, partial, start_idx):
        return [SelectableRow(self.non_filtered_columns, obj, start_idx + idx, on_select=self.on_select) if isinstance(obj, TaskRow) else ProjectPlaceholderRow(self.columns, obj, start_idx + idx) for idx, obj in enumerate(partial)]

    def build_table(self):
        self.make_header()
        self.batcher = ListBatcher(self.rows, self.list_walker, batch_to_formatter=self.format_task_batch)
        _, rows = self.listbox.size
        self.batcher.add(rows)

    def make_header(self):
        columns = []
        if len(self.columns) > 0:
            last_column = self.columns[-1]
            columns = [self.make_header_column(column, column == last_column) for column in self.columns]
        columns.append(self.make_padding('list-header-column'))
        list_header = urwid.Columns(columns)
        self.header = urwid.AttrMap(list_header, 'list-header')

    def make_header_column(self, column, is_last, space_between=COLUMN_PADDING):
        padding_width = 0 if is_last else space_between
        total_width = column['width'] + padding_width
        column_content = urwid.AttrMap(urwid.Padding(urwid.Text(column['label'], align='left')), 'list-header-column')
        padding_content = self.make_padding(is_last and 'list-header-column' or 'list-header-column-separator')
        columns = urwid.Columns([(column['width'], column_content), (padding_width, padding_content)])
        return (total_width, columns)

    def make_padding(self, display_attr):
        return urwid.AttrMap(urwid.Padding(urwid.Text('')), display_attr)

    def rows_size_grew(self, data):
        _, old_rows = data['old_size']
        _, new_rows = data['new_size']
        if new_rows > old_rows:
            return new_rows - old_rows
        return 0

    def size_changed(self, data):
        self.update_header(data['new_size'])
        grew = self.rows_size_grew(data)
        if grew > 0:
            self.batcher.add(grew)

class TaskRow:
    def __init__(self, task, data, alt_row):
        self.task = task
        self.data = data
        self.alt_row = alt_row
        self.uuid = self.task['uuid']
        self.id = self.task['id']

class ProjectRow:
    def __init__(self, project, placeholder, alt_row):
        self.project = project
        self.placeholder = placeholder
        self.alt_row = alt_row

class SelectableRow(urwid.WidgetWrap):
    """Wraps 'urwid.Columns' to make it selectable.
    This class has been slightly modified, but essentially corresponds to this class posted on stackoverflow.com:
    https://stackoverflow.com/questions/52106244/how-do-you-combine-multiple-tui-forms-to-write-more-complex-applications#answer-52174629"""

    def __init__(self, columns, row, position, *, on_select=None, space_between=COLUMN_PADDING):
        self.task = row.task
        self.uuid = row.uuid
        self.id = row.id
        self.alt_row = row.alt_row
        self.position = position

        self._columns = urwid.Columns([(column['width'], urwid.Text(row.data[idx], align=column['align'])) for idx, column in enumerate(columns) if column],
                                       dividechars=space_between)
        self.set_display_attr()
        self.row = urwid.AttrMap(self._columns, self.display_attr)

        # Wrap 'urwid.Columns'.
        super().__init__(self.row)

        # A hook which defines the behavior that is executed when a specified key is pressed.
        self.on_select = on_select

    def set_display_attr(self):
        self.display_attr = self.alt_row and 'striped-table-row' or ''

    def reset_attr_map(self):
        self.row.set_attr_map({None: self.display_attr})

    def __repr__(self):
        return "{}(id={}, uuid={})".format(self.__class__.__name__,
                                          self.id, self.uuid)

    def selectable(self):
        return True

    def keypress(self, size, key):
        if self.on_select:
            key = self.on_select(self, size, key)
        return key

class ProjectPlaceholderRow(urwid.WidgetWrap):
    """Wraps 'urwid.Columns' for a project placeholder row.
    """

    def __init__(self, columns, row, position, space_between=COLUMN_PADDING):
        self.uuid = None
        self.id = None
        self.alt_row = row.alt_row
        self.project = row.project
        self.placeholder = row.placeholder
        self.position = position
        self._columns = urwid.Columns([(column['width'], urwid.Text(row.placeholder if isinstance(column['formatter'], ProjectFormatter) else '', align=column['align'])) for column in columns], dividechars=space_between)

        self.set_display_attr()
        self.row = urwid.AttrMap(self._columns, self.display_attr)

        # Wrap 'urwid.Columns'.
        super().__init__(self.row)

    def set_display_attr(self):
        self.display_attr = self.alt_row and 'striped-table-row' or ''

    def reset_attr_map(self):
        self.row.set_attr_map({None: self.display_attr})
        pass

    def __repr__(self):
        return "{}(placeholder={})".format(self.__class__.__name__, self.placeholder)

class TaskListBox(BaseListBox):
    """Maps task list shortcuts to default ListBox class.
    """

    def __init__(self, body, screen, event=None, request_reply=None, action_manager=None):
        # TODO: Any way to get the actual listbox size here? It doesn't seem
        # to be accessible before the first render() call.
        self.screen = screen
        self.size = self.screen.get_cols_rows()
        return super().__init__(body, event, request_reply, action_manager)

    def render(self, size, focus=False):
        if size != self.size:
            data = {
                'old_size': self.size,
                'new_size': size,
            }
            self.size = size
            self.event.emit('task-list:size:change', data)
        return super().render(size, focus)

    def keypress_down(self, size):
        self.event.emit('task-list:keypress:down', size)
        super().keypress_down(size)

    def keypress_page_down(self, size):
        self.event.emit('task-list:keypress:page_down', size)
        super().keypress_page_down(size)

    def keypress_end(self, size):
        self.event.emit('task-list:keypress:end', size)
        super().keypress_end(size)

    def keypress_focus_valign_center(self, size):
        self.event.emit('task-list:keypress:focus_valign_center', size)
        super().keypress_focus_valign_center(size)

    def set_focus_position(self, start_idx=0):
        for idx, widget in enumerate(self.body[start_idx:]):
            if widget.selectable():
                self.set_focus(start_idx + idx)
                return

    def focus_by_batch(self, match_callback, start_idx):
        end_idx = start_idx
        for idx, row in enumerate(self.body[start_idx:]):
            end_idx = idx + start_idx
            if match_callback(row):
                self.focus_position = end_idx
                return True, end_idx
        return False, end_idx

    def focus_by_batch_loop(self, match_callback, previous_idx=0):
        start_idx = 0
        while True:
            found, start_idx = self.focus_by_batch(match_callback, start_idx)
            complete = self.request_reply.request('task-table:batch:next')
            if found:
                return
            elif complete:
                found, end_idx = self.focus_by_batch(match_callback, start_idx)
                if not found:
                    self.set_focus_position(end_idx if previous_idx > end_idx else previous_idx)
                return

    def focus_by_task_id(self, task_id):
        def match_callback(row):
            return row.id == task_id
        self.focus_by_batch_loop(match_callback)

    def focus_by_task_uuid(self, uuid, previous_idx=0):
        def match_callback(row):
            return row.uuid == uuid
        self.focus_by_batch_loop(match_callback, previous_idx)

    def list_action_executed(self, size, key):
        data = {
            'size': size,
        }
        self.event.emit('task-list:keypress', data)



import re

import urwid

from vit.base_list_box import BaseListBox
from vit.util import unicode_len

CURLY_BRACES_REGEX = re.compile("[{}]")
SPECIAL_KEY_SUBSTITUTIONS = {
    '<Colon>': ':',
    '<Equals>': '=',
}

class SelectableHelpRow(urwid.WidgetWrap):
    """Wraps 'urwid.Columns' to make it selectable.
    This class has been slightly modified, but essentially corresponds to this class posted on stackoverflow.com:
    https://stackoverflow.com/questions/52106244/how-do-you-combine-multiple-tui-forms-to-write-more-complex-applications#answer-52174629"""

    def __init__(self, column_widths, data, position, *, space_between=2):
        self.data = data
        self.position = position
        self._columns = urwid.Columns([
            (column_widths['type'], urwid.Text(self.data[0], align='left')),
            (column_widths['keys'], urwid.Text(self.data[1], align='left')),
            urwid.Text(self.data[2], align='left'),
        ], dividechars=space_between)
        self.row = urwid.AttrMap(self._columns, '', 'reveal focus')

        # Wrap 'urwid.Columns'.
        super().__init__(self.row)

    def __repr__(self):
        return "{}(type={}, keys={}, desc={})".format(self.__class__.__name__, self.data[0], self.data[1], self.data[2])

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key

class HelpListBox(BaseListBox):
    """Custom ListBox class for help.
    """

    def __init__(self, body, keybindings, event=None, request_reply=None, action_manager=None):
        self.keybindings = keybindings
        return super().__init__(body, event=event, request_reply=request_reply, action_manager=action_manager)

    def register_managed_actions(self):
        super().register_managed_actions()
        self.action_manager_registrar.register('GLOBAL_ESCAPE', self.exit_help)
        self.action_manager_registrar.register('QUIT_WITH_CONFIRM', self.exit_help)
        self.action_manager_registrar.register('QUIT', self.exit_help)

    def calculate_column_widths(self, entries):
        column_widths = {
            'type': 0,
            'keys': 0,
        }
        for entry in entries:
            type_len = unicode_len(entry[0])
            keys_len = unicode_len(entry[1])
            if type_len > column_widths['type']:
                column_widths['type'] = type_len
            if keys_len > column_widths['keys']:
                column_widths['keys'] = keys_len
        return column_widths

    def reload_entries(self, entries):
        column_widths = self.calculate_column_widths(entries)
        rows = [SelectableHelpRow(column_widths, row, idx) for idx, row in enumerate(entries)]
        self.list_walker[:] = rows
        if len(self.list_walker) > 0:
            self.set_focus(0)

    def exit_help(self, data):
        self.event.emit("help:exit")

    def eat_other_keybindings(self):
        return True

class Help:
    """Generates help list/display.
    """
    def __init__(self, keybinding_parser, actions, event=None, request_reply=None, action_manager=None):
        self.keybinding_parser = keybinding_parser
        self.actions = actions
        self.event = event
        self.request_reply = request_reply
        self.action_manager = action_manager
        self.keybindings = self.keybinding_parser.get_keybindings()
        sections = self.build_default_keybinding_data()
        sections = self.add_custom_help(sections)
        self.compose_entries(sections)
        self.build_help_widget()

    def autocomplete_entries(self):
        return [
            'help',
            'help command',
            'help global',
            'help help',
            'help navigation',
            'help report',
        ]

    def update(self, filter_args):
        entries = self.filter_entries(filter_args)
        self.listbox.reload_entries(entries)
        return self.widget

    def filter_entries(self, filter_args):
        if len(filter_args) > 0:
            args_regex = re.compile('.*(%s).*' % '|'.join(filter_args))
            return [(section, keys, description) for section, keys, description, search_phrase in self.entries if args_regex.match(search_phrase)]
        else:
            return [(section, keys, description) for section, keys, description, _ in self.entries]

    def compose_entries(self, sections):
        self.entries = []
        for section in self.keybinding_parser.sections:
            for keys, description in sections[section]:
                self.add_entry(section, keys, description)
        for keys, description in sections['help']:
            self.add_entry('help', keys, description)

    def add_entry(self, section, keys, description):
        self.entries.append((section, keys, description, ' '.join([section, keys, description])))

    # TODO: This does not pull data from keybinding overrides in config.ini
    # yet.
    def build_default_keybinding_data(self):
        sections = {}
        for section in self.keybinding_parser.sections:
            sections[section] = []
            for keys, action in self.keybinding_parser.items(section):
                action = re.sub(CURLY_BRACES_REGEX, '', action)
                keys = self.special_key_substitutions(keys)
                if action in self.actions:
                    sections[section].append((keys, self.actions[action]['description']))
        return sections

    def add_custom_help(self, sections):
        sections['command'] += [
            (':q', 'Quit the application'),
            (':![rw] STRING', "Execute STRING in shell. 'r' re-reads, 'w' waits"),
            (':s/OLD/NEW/', "Change OLD to NEW in the current task's description"),
            (':%s/OLD/NEW/', "Change OLD to NEW in the current task's description"),
        ]
        sections['navigation'] += [
            (':N', 'Move to task number N'),
        ]
        sections['report'] += [
            (':REPORT', 'Display REPORT (supports tab completion)'),
            (':REPORT FILTER', 'Display REPORT with FILTER (supports tab completion)'),
        ]
        sections['help'] = [
            (':help', 'View the whole help file'),
            (':help command', 'View help about commands'),
            (':help global', 'View help global actions'),
            (':help help', 'View help about help'),
            (':help navigation', 'View help about navigation'),
            (':help report', 'View help about reports'),
            (':help PATTERN', 'View help file lines matching PATTERN'),
        ]
        return sections

    def special_key_substitutions(self, keys):
        for key, sub in SPECIAL_KEY_SUBSTITUTIONS.items():
            keys = keys.replace(key, sub)
        return keys

    def build_help_widget(self):
        self.listbox = HelpListBox(urwid.SimpleFocusListWalker([]), self.keybinding_parser.keybindings, event=self.event, request_reply=self.request_reply, action_manager=self.action_manager)
        self.widget = urwid.Frame(
            self.listbox,
            header=urwid.Pile([
                urwid.Text("Press '<Esc>' or 'q' to exit", align='center'),
                # TODO: Remove this when keybinding overrides are shown.
                urwid.Text("NOTE: Keybinding overrides in config.ini not shown", align='center'),
                urwid.Text(''),
            ]),
        )



class Actions:

    def __init__(self, action_registry):
        self.action_registry = action_registry

    def register(self):
        self.action_registrar = self.action_registry.get_registrar()
        # Global.
        self.action_registrar.register('QUIT', 'Quit the application')
        self.action_registrar.register('QUIT_WITH_CONFIRM', 'Quit the application, after confirmation')
        self.action_registrar.register('GLOBAL_ESCAPE', 'Top-level escape function')
        self.action_registrar.register('REFRESH', 'Refresh the current report')
        self.action_registrar.register('TASK_ADD', 'Add a task (supports tab completion)')
        self.action_registrar.register('REPORT_FILTER', 'Filter current report using provided FILTER arguments (supports tab completion)')
        self.action_registrar.register('TASK_UNDO', 'Undo last task change')
        self.action_registrar.register('TASK_SYNC', 'Synchronize with configured taskd server')
        self.action_registrar.register('COMMAND_BAR_EX', "Open the command bar in 'ex' mode")
        self.action_registrar.register('COMMAND_BAR_EX_TASK_READ_WAIT', "Open the command bar in 'ex' mode with '!rw task ' appended")
        self.action_registrar.register('COMMAND_BAR_SEARCH_FORWARD', 'Search forward for provided STRING')
        self.action_registrar.register('COMMAND_BAR_SEARCH_REVERSE', 'Search reverse for provided STRING')
        self.action_registrar.register('COMMAND_BAR_SEARCH_NEXT', 'Search next')
        self.action_registrar.register('COMMAND_BAR_SEARCH_PREVIOUS', 'Search previous')
        self.action_registrar.register('COMMAND_BAR_TASK_CONTEXT', 'Set task context')
        self.action_registrar.register(self.action_registry.noop_action_name, 'Used to disable a default keybinding action')
        # List.
        self.action_registrar.register('LIST_UP', 'Move list focus up one entry')
        self.action_registrar.register('LIST_DOWN', 'Move list focus down one entry')
        self.action_registrar.register('LIST_PAGE_UP', 'Move list focus up one page')
        self.action_registrar.register('LIST_PAGE_DOWN', 'Move list focus down one page')
        self.action_registrar.register('LIST_HOME', 'Move list focus to top of the list')
        self.action_registrar.register('LIST_END', 'Move list focus to bottom of the list')
        self.action_registrar.register('LIST_SCREEN_TOP', 'Move list focus to top of the screen')
        self.action_registrar.register('LIST_SCREEN_MIDDLE', 'Move list focus to middle of the screen')
        self.action_registrar.register('LIST_SCREEN_BOTTOM', 'Move list focus to bottom of the screen')
        self.action_registrar.register('LIST_FOCUS_VALIGN_CENTER', 'Move focused item to center of the screen')
        # Task.
        self.action_registrar.register('TASK_ANNOTATE', 'Add an annotation to a task')
        self.action_registrar.register('TASK_DELETE', 'Delete task')
        self.action_registrar.register('TASK_DENOTATE', 'Denotate a task')
        self.action_registrar.register('TASK_MODIFY', 'Modify task (supports tab completion)')
        self.action_registrar.register('TASK_START_STOP', 'Start/stop task')
        self.action_registrar.register('TASK_DONE', 'Mark task done')
        self.action_registrar.register('TASK_PRIORITY', 'Modify task priority')
        self.action_registrar.register('TASK_PROJECT', 'Modify task project (supports tab completion)')
        self.action_registrar.register('TASK_TAGS', 'Modify task tags (supports tab completion, +TAG adds, -TAG removes)')
        self.action_registrar.register('TASK_WAIT', 'Wait a task')
        self.action_registrar.register('TASK_EDIT', 'Edit a task via the default editor')
        self.action_registrar.register('TASK_SHOW', 'Show task details')

    def get(self):
        return self.action_registry.actions



from vit.formatter.description import Description
from vit.util import unicode_len

class DescriptionTruncated(Description):
    def format(self, description, task):
        if not description:
            return self.empty()
        truncated_description = self.format_description_truncated(description)
        width = unicode_len(truncated_description)
        colorized_description = self.colorize_description(truncated_description)
        return (width, colorized_description)
